@@ -42,11 +42,12 @@ You cannot use this program to actually change your UID.\n";
 
 const char args_doc[] = "IMAGE -- CMD [ARG...]";
 
+/* Note: Long option numbers, once issued, are permanent; i.e., if you remove
+   one, don’t re-number the others. */
 const struct argp_option options[] = {
    { "bind",          'b', "SRC[:DST]", 0,
      "mount SRC at guest DST (default: same as SRC)"},
    { "cd",            'c', "DIR",  0, "initial working directory in container"},
-   { "ch-ssh",         -8, 0,      0, "bind ch-ssh into image"},
    { "env-no-expand", -10, 0,      0, "don't expand $ in --set-env input"},
    { "feature",       -11, "FEAT", 0, "exit successfully if FEAT is enabled" },
    { "gid",           'g', "GID",  0, "run as GID within container" },
@@ -56,11 +57,14 @@ const struct argp_option options[] = {
    { "join-ct",        -3, "N",    0, "number of join peers (implies --join)" },
    { "join-tag",       -4, "TAG",  0, "label for peer group (implies --join)" },
    { "mount",         'm', "DIR",  0, "SquashFS mount point"},
-   { "no-home",        -2, 0,      0, "(deprecated)"},
    { "no-passwd",      -9, 0,      0, "don't bind-mount /etc/{passwd,group}"},
    { "private-tmp",   't', 0,      0, "use container-private /tmp" },
+#ifdef HAVE_SECCOMP
+   { "seccomp",       -14, 0,      0,
+                           "fake success for some syscalls with seccomp(2)"},
+#endif
    { "set-env",        -6, "ARG",  OPTION_ARG_OPTIONAL,
-     "set environment variables per ARG"},
+                           "set environment variables per ARG"},
    { "storage",       's', "DIR",  0, "set DIR as storage directory"},
    { "uid",           'u', "UID",  0, "run as UID within container" },
    { "unsafe",        -13, 0,      0, "do unsafe things (internal use only)" },
@@ -77,8 +81,11 @@ const struct argp_option options[] = {
 struct args {
    struct container c;
    struct env_delta *env_deltas;
-   char *storage_dir;
    char *initial_dir;
+#ifdef HAVE_SECCOMP
+   bool seccomp_p;
+#endif
+   char *storage_dir;
    bool unsafe;
 };
 
@@ -124,7 +131,6 @@ int main(int argc, char *argv[])
    verbose = LL_INFO;  // in ch_misc.c
    args = (struct args){
       .c = (struct container){ .binds = list_new(sizeof(struct bind), 0),
-                               .ch_ssh = false,
                                .container_gid = getegid(),
                                .container_uid = geteuid(),
                                .env_expand = true,
@@ -140,8 +146,11 @@ int main(int argc, char *argv[])
                                .type = IMG_NONE,
                                .writable = false },
       .env_deltas = list_new(sizeof(struct env_delta), 0),
-      .storage_dir = storage_default(),
       .initial_dir = NULL,
+#ifdef HAVE_SECCOMP
+      .seccomp_p = false,
+#endif
+      .storage_dir = storage_default(),
       .unsafe = false };
 
    /* I couldn't find a way to set argp help defaults other than this
@@ -207,9 +216,17 @@ int main(int argc, char *argv[])
    VERBOSE("join: %d %d %s %d", args.c.join, args.c.join_ct, args.c.join_tag,
            args.c.join_pid);
    VERBOSE("private /tmp: %d", args.c.private_tmp);
+#ifdef HAVE_SECCOMP
+   VERBOSE("seccomp: %d", args.seccomp_p);
+#endif
+   VERBOSE("unsafe: %d", args.unsafe);
 
    containerize(&args.c);
    fix_environment(&args);
+#ifdef HAVE_SECCOMP
+   if (args.seccomp_p)
+      seccomp_install();
+#endif
    run_user_command(c_argv, args.initial_dir); // should never return
    exit(EXIT_FAILURE);
 }
@@ -368,9 +385,6 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
    int i;
 
    switch (key) {
-   case -2: // --no-home
-      WARNING("deprecated --no-home is now default; ignoring")
-      break;
    case -3: // --join-ct
       args->c.join = true;
       args->c.join_ct = parse_int(arg, false, "--join-ct");
@@ -402,9 +416,6 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       ed.arg.glob = arg;
       list_append((void **)&(args->env_deltas), &ed, sizeof(ed));
       break;;
-   case -8: // --ch-ssh
-      args->c.ch_ssh = true;
-      break;
    case -9: // --no-passwd
       args->c.private_passwd = true;
       break;
@@ -418,7 +429,20 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
 #else
          exit(1);
 #endif
-      } else
+      } else if (!strcmp(arg, "seccomp")) {
+#ifdef HAVE_SECCOMP
+         exit(0);
+#else
+         exit(1);
+#endif
+      } else if (!strcmp(arg, "squash")) {
+#ifdef HAVE_LIBSQUASHFUSE
+         exit(0);
+#else
+         exit(1);
+#endif
+      }
+      else
          FATAL("unknown feature: %s", arg);
       break;
    case -12: // --home
@@ -427,6 +451,11 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
    case -13: // --unsafe
       args->unsafe = true;
       break;
+#ifdef HAVE_SECCOMP
+   case -14: // --seccomp
+      args->seccomp_p = true;
+      break;
+#endif
    case 'b': {  // --bind
          char *src, *dst;
          for (i = 0; args->c.binds[i].src != NULL; i++) // count existing binds
