@@ -29,7 +29,6 @@ enum img_type {
 
 struct container {
    struct bind *binds;
-   bool ch_ssh;          // bind /usr/bin/ch-ssh?
    gid_t container_gid;  // GID to use in container
    uid_t container_uid;  // UID to use in container
    bool env_expand;      // expand variables in --set-env
@@ -53,3 +52,6 @@ void containerize(struct container *c);
 enum img_type image_type(const char *ref, const char *images_dir);
 char *img_name2path(const char *name, const char *storage_dir);
 void run_user_command(char *argv[], const char *initial_dir);
+#ifdef HAVE_SECCOMP
+void seccomp_install(void);
+#endif
