@@ -13,32 +13,32 @@ import sys
 
 import charliecloud as ch
 import build_cache as bu
-import image as im
-import fakeroot
-import pull
 import filesystem as fs
+import force
+import image as im
+import pull
 
 
 ## Globals ##
 
+# ARG values that are set before FROM.
+argfrom = {}
+
 # Namespace from command line arguments. FIXME: be more tidy about this ...
 cli = None
 
-# Fakeroot configuration (initialized during FROM).
-fakeroot_config = None
+# --force injector object (initialized to something meaningful during FROM).
+forcer = None
 
 # Images that we are building. Each stage gets its own image. In this
 # dictionary, an image appears exactly once or twice. All images appear with
-# an int key counting stages up from zero. Images with a name (e.g., "FROM ...
-# AS foo") have a second string key of the name.
+# an int key counting stages up from zero. Images with a name (e.g., “FROM ...
+# AS foo”) have a second string key of the name.
 images = dict()
 # Number of stages. This is obtained by counting FROM instructions in the
 # parse tree, so we can use it for error checking.
 image_ct = None
 
-# ARG values that are set before FROM.
-argfrom = {}
-
 
 ## Imports not in standard library ##
 
@@ -59,10 +59,6 @@ def main(cli_):
    global cli
    cli = cli_
 
-   # Check argument validity.
-   if (cli.force and cli.no_force_detect):
-      ch.FATAL("--force and --no-force-detect are incompatible")
-
    # Infer input file if needed.
    if (cli.file is None):
       cli.file = cli.context + "/Dockerfile"
@@ -92,6 +88,28 @@ def main(cli_):
       cli.tag = re.sub(r"[^a-z0-9_.-]", "", cli.tag.lower())
       ch.INFO("inferred image name: %s" % cli.tag)
 
+   # --force and friends.
+   if (cli.force_cmd and cli.force == "fakeroot"):
+      ch.FATAL("--force-cmd and --force=fakeroot are incompatible")
+   if (not cli.force_cmd):
+      cli.force_cmd = force.FORCE_CMD_DEFAULT
+   else:
+      cli.force = "seccomp"
+      # convert cli.force_cmd to parsed dict
+      force_cmd = dict()
+      for line in cli.force_cmd:
+         (cmd, args) = force.force_cmd_parse(line)
+         force_cmd[cmd] = args
+      cli.force_cmd = force_cmd
+   ch.VERBOSE("force mode: %s" % cli.force)
+   if (cli.force == "seccomp"):
+      for (cmd, args) in cli.force_cmd.items():
+         ch.VERBOSE("force command: %s" % ch.argv_to_string([cmd] + args))
+   if (    cli.force == "seccomp"
+       and ch.cmd([ch.CH_BIN + "/ch-run", "--feature=seccomp"],
+                  fail_ok=True) != 0):
+      ch.FATAL("ch-run was not built with seccomp(2) support")
+
    # Deal with build arguments.
    def build_arg_get(arg):
       kv = arg.split("=")
@@ -106,8 +124,8 @@ def main(cli_):
    ch.DEBUG(cli)
 
    # Guess whether the context is a URL, and error out if so. This can be a
-   # typical looking URL e.g. "https://..." or also something like
-   # "git@github.com:...". The line noise in the second line of the regex is
+   # typical looking URL e.g. “https://...” or also something like
+   # “git@github.com:...”. The line noise in the second line of the regex is
    # to match this second form. Username and host characters from
    # https://tools.ietf.org/html/rfc3986.
    if (re.search(r"""  ^((git|git+ssh|http|https|ssh)://
@@ -119,12 +137,12 @@ def main(cli_):
 
    # Read input file.
    if (cli.file == "-" or cli.context == "-"):
-      text = ch.ossafe(sys.stdin.read, "can't read stdin")
+      text = ch.ossafe(sys.stdin.read, "can’t read stdin")
    elif (not os.path.isdir(cli.context)):
       ch.FATAL("context must be a directory: %s" % cli.context)
    else:
       fp = fs.Path(cli.file).open_("rt")
-      text = ch.ossafe(fp.read, "can't read: %s" % cli.file)
+      text = ch.ossafe(fp.read, "can’t read: %s" % cli.file)
       ch.close_(fp)
 
    # Parse it.
@@ -137,7 +155,7 @@ def main(cli_):
       tree = parser.parse(text)
    except lark.exceptions.UnexpectedInput as x:
       ch.VERBOSE(x)  # noise about what was expected in the grammar
-      ch.FATAL("can't parse: %s:%d,%d\n\n%s"
+      ch.FATAL("can’t parse: %s:%d,%d\n\n%s"
                % (cli.file, x.line, x.column, x.get_context(text, 39)))
    ch.VERBOSE(tree.pretty())
 
@@ -151,7 +169,7 @@ def main(cli_):
 
    # Traverse the tree and do what it says.
    #
-   # We don't actually care whether the tree is traversed breadth-first or
+   # We don’t actually care whether the tree is traversed breadth-first or
    # depth-first, but we *do* care that instruction nodes are visited in
    # order. Neither visit() nor visit_topdown() are documented as of
    # 2020-06-11 [1], but examining source code [2] shows that visit_topdown()
@@ -178,20 +196,16 @@ def main(cli_):
    if (len(cli.build_arg) != 0):
       ch.FATAL("--build-arg: not consumed: " + " ".join(cli.build_arg.keys()))
 
-   # Print summary & we're done.
+   # Print summary & we’re done.
    if (ml.instruction_total_ct == 0):
       ch.FATAL("no instructions found: %s" % cli.file)
    assert (ml.inst_prev.image_i + 1 == image_ct)  # should’ve errored already
    if (cli.force and ml.miss_ct != 0):
-      if (fakeroot_config.inject_ct == 0):
-         assert (not fakeroot_config.init_done)
-         ch.WARNING("--force specified, but nothing to do")
-      else:
-         ch.INFO("--force: init OK & modified %d RUN instructions"
-                 % fakeroot_config.inject_ct)
+      ch.INFO("--force=%s: modified %d RUN instructions"
+              % (cli.force, forcer.run_modified_ct))
    ch.INFO("grown in %d instructions: %s"
            % (ml.instruction_total_ct, ml.inst_prev.image))
-   # FIXME: remove when we're done encouraging people to use the build cache.
+   # FIXME: remove when we’re done encouraging people to use the build cache.
    if (isinstance(bu.cache, bu.Disabled_Cache)):
       ch.INFO("build slow? consider enabling the new build cache",
               "https://hpc.github.io/charliecloud/command-usage.html#build-cache")
@@ -264,7 +278,7 @@ class Instruction(abc.ABC):
 
    def __init__(self, tree):
       """Note: When this is called, all we know about the instruction is
-         what's in the parse tree. In particular, you must not call
+         what’s in the parse tree. In particular, you must not call
          ch.variables_sub() here."""
       self.announced_p = False
       self.commit_files = set()
@@ -392,9 +406,8 @@ class Instruction(abc.ABC):
 
    def checkout_for_build(self, base_image=None):
       self.parent.checkout(base_image)
-      global fakeroot_config
-      fakeroot_config = fakeroot.detect(self.image.unpack_path,
-                                        cli.force, cli.no_force_detect)
+      global forcer
+      forcer = force.new(self.image.unpack_path, cli.force, cli.force_cmd)
 
    def commit(self):
       path = self.image.unpack_path
@@ -410,13 +423,13 @@ class Instruction(abc.ABC):
       pass
 
    def init(self, parent):
-      """Initialize attributes defining this instruction's context, much of
+      """Initialize attributes defining this instruction’s context, much of
          which is not available until the previous instruction is processed.
          After this is called, the instruction has a valid image and parent
-         instruction, unless it's the first instruction, in which case
+         instruction, unless it’s the first instruction, in which case
          prepare() does the initialization."""
-      # Separate from prepare() because subclasses shouldn't need to override
-      # it. If a subclass doesn't like the result, it can just change things
+      # Separate from prepare() because subclasses shouldn’t need to override
+      # it. If a subclass doesn’t like the result, it can just change things
       # in prepare().
       self.parent = parent
       if (self.parent is None):
@@ -463,8 +476,9 @@ class Instruction(abc.ABC):
               announces, and then re-raises.
 
            3. Modifying image metadata: Instructions like ARG, ENV, FROM,
-              SHELL, and WORKDIR must modify metadata here, not in execute(),
-              so it’s available to later instructions even on cache hit."""
+              LABEL, SHELL, and WORKDIR must modify metadata here, not in
+              execute(), so it’s available to later instructions even on
+              cache hit."""
       self.sid = bu.cache.sid_from_parent(self.parent.sid, self.sid_input)
       self.git_hash = bu.cache.find_sid(self.sid, self.image.ref.for_path)
       return miss_ct + int(self.miss)
@@ -650,7 +664,7 @@ class I_copy(Instruction):
    # bug-compatible with Docker but probably are not 100%. See the FAQ.
    #
    # Because of these weird semantics, none of this abstracted into a general
-   # copy function. I don't want people calling it except from here.
+   # copy function. I don’t want people calling it except from here.
 
    __slots__ = ("dst",
                 "dst_raw",
@@ -696,7 +710,7 @@ class I_copy(Instruction):
          must exist already and be a directory. Unlike subdirectories, the
          metadata of dst will not be altered to match src."""
       def onerror(x):
-         ch.FATAL("can't scan directory: %s: %s" % (x.filename, x.strerror))
+         ch.FATAL("can’t scan directory: %s: %s" % (x.filename, x.strerror))
       # Use Path objects in this method because the path arithmetic was
       # getting too hard with strings.
       src = src.resolve()  # alternative to os.path.realpath()
@@ -707,8 +721,8 @@ class I_copy(Instruction):
       for (dirpath, dirnames, filenames) in ch.walk(src, onerror=onerror):
          subdir = dirpath.relative_to(src)
          dst_dir = dst // subdir
-         # dirnames can contain symlinks, which we handle as files, so we'll
-         # rebuild it; the walk will not descend into those "directories".
+         # dirnames can contain symlinks, which we handle as files, so we’ll
+         # rebuild it; the walk will not descend into those “directories”.
          dirnames2 = dirnames.copy()  # shallow copy
          dirnames[:] = list()         # clear in place
          for d in dirnames2:
@@ -721,21 +735,21 @@ class I_copy(Instruction):
                continue
             else:
                dirnames.append(d)   # directory, descend into later
-            # If destination exists, but isn't a directory, remove it.
+            # If destination exists, but isn’t a directory, remove it.
             if (os.path.exists(dst_path)):
                if (os.path.isdir(dst_path) and not os.path.islink(dst_path)):
                   ch.TRACE("dst_path exists and is a directory")
                else:
                   ch.TRACE("dst_path exists, not a directory, removing")
                   dst_path.unlink_()
-            # If destination directory doesn't exist, create it.
+            # If destination directory doesn’t exist, create it.
             if (not os.path.exists(dst_path)):
                ch.TRACE("mkdir dst_path")
-               ch.ossafe(os.mkdir, "can't mkdir: %s" % dst_path, dst_path)
+               ch.ossafe(os.mkdir, "can’t mkdir: %s" % dst_path, dst_path)
             # Copy metadata, now that we know the destination exists and is a
             # directory.
             ch.ossafe(shutil.copystat,
-                      "can't copy metadata: %s -> %s" % (src_path, dst_path),
+                      "can’t copy metadata: %s -> %s" % (src_path, dst_path),
                       src_path, dst_path, follow_symlinks=False)
          for f in filenames:
             src_path = dirpath // f
@@ -743,7 +757,7 @@ class I_copy(Instruction):
             ch.TRACE("file or symlink via copy2: %s -> %s"
                       % (src_path, dst_path))
             if (not (os.path.isfile(src_path) or os.path.islink(src_path))):
-               ch.FATAL("can't COPY: unknown file type: %s" % src_path)
+               ch.FATAL("can’t COPY: unknown file type: %s" % src_path)
             if (os.path.exists(dst_path)):
                ch.TRACE("destination exists, removing")
                if (os.path.isdir(dst_path) and not os.path.islink(dst_path)):
@@ -777,8 +791,8 @@ class I_copy(Instruction):
 
    def dest_realpath(self, unpack_path, dst):
       """Return the canonicalized version of path dst within (canonical) image
-         path unpack_path. We can't use os.path.realpath() because if dst is
-         an absolute symlink, we need to use the *image's* root directory, not
+         path unpack_path. We can’t use os.path.realpath() because if dst is
+         an absolute symlink, we need to use the *image’s* root directory, not
          the host. Thus, we have to resolve symlinks manually."""
       dst_canon = unpack_path
       dst_parts = list(reversed(dst.parts))  # easier to operate on end of list
@@ -786,7 +800,7 @@ class I_copy(Instruction):
       while (len(dst_parts) > 0):
          iter_ct += 1
          if (iter_ct > 100):  # arbitrary
-            ch.FATAL("can't COPY: too many path components")
+            ch.FATAL("can’t COPY: too many path components")
          ch.TRACE("current destination: %d %s" % (iter_ct, dst_canon))
          #ch.TRACE("parts remaining: %s" % dst_parts)
          part = dst_parts.pop()
@@ -822,7 +836,7 @@ class I_copy(Instruction):
       ch.VERBOSE("destination, canonical: %s" % dst_canon)
       if (not os.path.commonpath([dst_canon, unpack_canon])
               .startswith(str(unpack_canon))):
-         ch.FATAL("can't COPY: destination not in image: %s" % dst_canon)
+         ch.FATAL("can’t COPY: destination not in image: %s" % dst_canon)
       # Create the destination directory if needed.
       if (   self.dst.endswith("/")
           or len(self.srcs) > 1
@@ -830,7 +844,7 @@ class I_copy(Instruction):
          if (not dst_canon.exists()):
             dst_canon.mkdirs()
          elif (not dst_canon.is_dir()):  # not symlink b/c realpath()
-            ch.FATAL("can't COPY: not a directory: %s" % dst_canon)
+            ch.FATAL("can’t COPY: not a directory: %s" % dst_canon)
       if (dst_canon.parent.exists()):
          if (not dst_canon.parent.is_dir()):
             ch.FATAL("can’t COPY: not a directory: %s" % dst_canon.parent)
@@ -843,7 +857,7 @@ class I_copy(Instruction):
          elif (src.is_dir()):
             self.copy_src_dir(src, dst_canon)
          else:
-            ch.FATAL("can't COPY: unknown file type: %s" % src)
+            ch.FATAL("can’t COPY: unknown file type: %s" % src)
 
    def prepare(self, miss_ct):
       def stat_bytes(path, links=False):
@@ -852,7 +866,7 @@ class I_copy(Instruction):
                  + struct.pack("=HQQ", st.st_mode, st.st_size, st.st_mtime_ns))
       # Error checking.
       if (cli.context == "-" and self.from_ is None):
-         ch.FATAL("no context because \"-\" given")
+         ch.FATAL("no context because “-” given")
       if (len(self.srcs_raw) < 1):
          ch.FATAL("must specify at least one source")
       # Complain about unsupported stuff.
@@ -892,7 +906,7 @@ class I_copy(Instruction):
             ch.VERBOSE("source: %s" % i)
       # Expand destination.
       self.dst = ch.variables_sub(self.dst_raw, self.env_build)
-      # Validate sources are within context directory. (Can't convert to
+      # Validate sources are within context directory. (Can’t convert to
       # canonical paths yet because we need the source path as given.)
       for src in self.srcs:
          src_canon = src.resolve()
@@ -976,9 +990,51 @@ class I_env_space(Env):
       self.value = self.tree.terminals_cat("LINE_CHUNK")
 
 
+class Label(Instruction):
+
+   __slots__ = ("key",
+                "value")
+
+   def __init__(self, *args):
+      super().__init__(*args)
+      self.commit_files |= {ch.Path("ch/metadata.json")}
+
+   @property
+   def str_(self):
+      return "%s='%s'" % (self.key, self.value)
+
+   def prepare(self, *args):
+      self.value = ch.variables_sub(unescape(self.value), self.env_build)
+      self.image.metadata["labels"][self.key] = self.value
+      return super().prepare(*args)
+
+
+class I_label_equals(Label):
+
+   __slots__ = ()
+
+   def __init__(self, *args):
+      super().__init__(*args)
+      self.key = self.tree.terminal("WORD", 0)
+      self.value = self.tree.terminal("WORD", 1)
+      if (self.value is None):
+         self.value = self.tree.terminal("STRING_QUOTED")
+
+
+class I_label_space(Label):
+
+   __slots__ = ()
+
+   def __init__(self, *args):
+      super().__init__(*args)
+      self.key = self.tree.terminal("WORD")
+      self.value = self.tree.terminals_cat("LINE_CHUNK")
+
+
 class I_from_(Instruction):
 
    __slots__ = ("alias",
+                "base_alias",
                 "base_image",
                 "base_text")
 
@@ -991,12 +1047,14 @@ class I_from_(Instruction):
 
    @property
    def str_(self):
-      if (hasattr(self, "base_image")):
-         base_image = str(self.base_image.ref)
+      if (hasattr(self, "base_alias")):
+         base_text = str(self.base_alias)
+      elif (hasattr(self, "base_image")):
+         base_text = str(self.base_image.ref)
       else:
          # Initialization failed, but we want to print *something*.
-         base_image = self.base_text
-      return base_image + ((" AS " + self.alias) if self.alias else "")
+         base_text = self.base_text
+      return base_text + ((" AS " + self.alias) if self.alias else "")
 
    def checkout_for_build(self):
       assert (isinstance(bu.cache, bu.Disabled_Cache))
@@ -1014,7 +1072,6 @@ class I_from_(Instruction):
       # parent is the last instruction of the base image.
       self.base_text = self.tree.child_terminals_cat("image_ref", "IMAGE_REF")
       self.alias = self.tree.child_terminal("from_alias", "IR_PATH_COMPONENT")
-      self.base_image = im.Image(im.Reference(self.base_text, argfrom))
       # Validate instruction.
       if (self.options.pop("platform", False)):
          self.unsupported_yet_fatal("--platform", 778)
@@ -1032,6 +1089,12 @@ class I_from_(Instruction):
       else:
          # Not last image; append stage index to tag.
          tag = "%s_stage%d" % (cli.tag, self.image_i)
+      if self.base_text in images:
+         # Is alias; store base_text as the “alias used” to target a previous
+         # stage as the base.
+         self.base_alias = self.base_text
+         self.base_text = str(images[self.base_text].ref)
+      self.base_image = im.Image(im.Reference(self.base_text, argfrom))
       self.image = im.Image(im.Reference(tag))
       images[self.image_i] = self.image
       if (self.image_alias is not None):
@@ -1040,13 +1103,17 @@ class I_from_(Instruction):
       # More error checking.
       if (str(self.image.ref) == str(self.base_image.ref)):
          ch.FATAL("output image ref same as FROM: %s" % self.base_image.ref)
-      # Close previous stage if needed.
-      if (miss_ct == 0 and self.image_i > 0):
-         # While there haven't been any misses so far, we do need to check out
-         # the previous stage (a) to read its metadata and (b) in case there's
-         # a COPY later. This will still be fast most of the time since the
-         # correct branch is likely to be checked out already.
-         self.parent.checkout()
+      # Close previous stage if needed. In particular, we need the previous
+      # stage’s image directory to exist because (a) we need to read its
+      # metadata and (b) in case there’s a COPY later. Cache disabled will
+      # already have the image directory and there is no notion of branch
+      # “ready”, so do nothing in that case.
+      if (self.image_i > 0 and not isinstance(bu.cache, bu.Disabled_Cache)):
+         if (miss_ct == 0):
+            # No previous miss already checked out the image. This will still
+            # be fast most of the time since the correct branch is likely
+            # checked out already.
+            self.parent.checkout()
          self.parent.ready()
       # At this point any meaningful parent of FROM, e.g., previous stage, has
       # been closed; thus, act as own parent.
@@ -1085,34 +1152,33 @@ class Run(Instruction):
 
    __slots__ = ("cmd")
 
-   # FIXME: This causes spurious misses because it adds the force bit to *all*
-   # RUN instructions, not just those that actually were modified (i.e, any
-   # RUN instruction will miss the equivalent RUN with --force inverted). But
-   # we don't know know if an instruction needs modifications until the result
-   # is checked out, which happens after we check the cache. See issue #FIXME.
    @property
    def str_name(self):
-      return super().str_name + (".F" if cli.force else "")
+      # Can’t get this from the forcer object because it might not have been
+      # initialized yet.
+      if (cli.force is None):
+         tag = ""
+      elif (cli.force == "fakeroot"):
+         # FIXME: This causes spurious misses because it adds the force tag to
+         # *all* RUN instructions, not just those that actually were modified
+         # (i.e, any RUN instruction will miss the equivalent RUN without
+         # --force=fakeroot). But we don’t know know if an instruction needs
+         # modifications until the result is checked out, which happens after
+         # we check the cache. See issue #1339.
+         tag = ".F"
+      elif (cli.force == "seccomp"):
+         tag = ".S"
+      else:
+         assert False, "unreachable code reached"
+      return super().str_name + tag
 
    def execute(self):
       rootfs = self.image.unpack_path
-      fakeroot_config.init_maybe(rootfs, self.cmd, self.env_build)
-      cmd = fakeroot_config.inject_run(self.cmd)
+      cmd = forcer.run_modified(self.cmd, self.env_build)
       exit_code = ch.ch_run_modify(rootfs, cmd, self.env_build, self.workdir,
-                                   cli.bind, fail_ok=True)
+                                   cli.bind, forcer.ch_run_args, fail_ok=True)
       if (exit_code != 0):
-         msg = "build failed: RUN command exited with %d" % exit_code
-         if (cli.force):
-            if (isinstance(fakeroot_config, fakeroot.Fakeroot_Noop)):
-               ch.FATAL(msg, "--force specified, but no suitable config found")
-            else:
-               ch.FATAL(msg)  # --force inited OK but the build still failed
-         elif (not cli.no_force_detect):
-            if (fakeroot_config.init_done):
-               ch.FATAL(msg, "--force may fix it")
-            else:
-               ch.FATAL(msg, "current version of --force wouldn't help")
-         assert False, "unreachable code reached"
+         ch.FATAL("build failed: RUN command exited with %d" % exit_code)
 
 
 class I_run_exec(Run):
@@ -1139,7 +1205,7 @@ class I_run_shell(Run):
 
    @property
    def str_(self):
-      return self._str_  # can't replace abstract property with attribute
+      return self._str_  # can’t replace abstract property with attribute
 
    def prepare(self, *args):
       cmd = self.tree.terminals_cat("LINE_CHUNK")
@@ -1206,7 +1272,6 @@ class I_uns_yet(Instruction_Unsupported):
       self.issue_no = { "ADD":         782,
                         "CMD":         780,
                         "ENTRYPOINT":  780,
-                        "LABEL":       781,
                         "ONBUILD":     788 }[self.name]
 
    @property
@@ -1226,13 +1291,13 @@ class Environment:
 
    # FIXME:
    # - problem:
-   #   1. COPY (at least) needs a valid build environment to figure out if it's
+   #   1. COPY (at least) needs a valid build environment to figure out if it’s
    #      a hit or miss, which happens in prepare()
-   #   2. no files from the image are available in prepare(), so we can't read
+   #   2. no files from the image are available in prepare(), so we can’t read
    #      image metadata then
    #      - could get it from Git if needed, but that seems complicated
    # - valid during prepare() and execute() but not __init__()
-   #   - in particular, don't ch.variables_sub() in __init__()
+   #   - in particular, don’t ch.variables_sub() in __init__()
    # - instructions that update it need to change the env object in prepare()
    #   - WORKDIR SHELL ARG ENV
    #   - FROM
@@ -1247,8 +1312,8 @@ class Environment:
 def unescape(sl):
    # FIXME: This is also ugly and should go in the grammar.
    #
-   # The Dockerfile spec does not precisely define string escaping, but I'm
-   # guessing it's the Go rules. You will note that we are using Python rules.
+   # The Dockerfile spec does not precisely define string escaping, but I’m
+   # guessing it’s the Go rules. You will note that we are using Python rules.
    # This is wrong but close enough for now (see also gripe in previous
    # paragraph).
    if (    not sl.startswith('"')                          # no start quote
