@@ -4,13 +4,14 @@ import sys
 
 import charliecloud as ch
 import build_cache as bu
+import filesystem as fs
 import image as im
 import registry as rg
 
 
 ## Constants ##
 
-# Internal library of manifests, e.g. for "FROM scratch" (issue #1013).
+# Internal library of manifests, e.g. for “FROM scratch” (issue #1013).
 manifests_internal = {
    "scratch": {  # magic empty image
       "schemaVersion": 2,
@@ -90,6 +91,7 @@ class Image_Puller:
       "Download image metadata and layers and put them in the download cache."
       # Spec: https://docs.docker.com/registry/spec/manifest-v2-2/
       ch.VERBOSE("downloading image: %s" % self.image)
+      have_skinny = False
       try:
          # fat manifest
          if (ch.arch != "yolo"):
@@ -100,8 +102,11 @@ class Image_Puller:
                            ("available: %s"
                             % " ".join(sorted(self.architectures.keys()))))
             except ch.No_Fatman_Error:
+               # currently, this error is only raised if we’ve downloaded the
+               # skinny manifest.
+               have_skinny = True
                if (ch.arch == "amd64"):
-                  # We're guessing that enough arch-unaware images are amd64 to
+                  # We’re guessing that enough arch-unaware images are amd64 to
                   # barge ahead if requested architecture is amd64.
                   ch.arch = "yolo"
                   ch.WARNING("image is architecture-unaware")
@@ -110,7 +115,7 @@ class Image_Puller:
                   ch.FATAL("image is architecture-unaware",
                            "consider --arch=yolo")
          # manifest
-         self.manifest_load()
+         self.manifest_load(have_skinny)
       except ch.Image_Unavailable_Error:
          if (ch.user() == "qwofford"):
             h = "Quincy, use --auth!!"
@@ -173,9 +178,14 @@ class Image_Puller:
                                    "manifest list: downloading")
       fm = self.fatman_path.json_from_file("fat manifest")
       if ("layers" in fm or "fsLayers" in fm):
-         # FIXME (issue #1101): If it's a v2 manifest we could use it instead
-         # of re-requesting later. Maybe we could here move/copy it over to
-         # the skinny manifest path.
+         # Check for skinny manifest. If not present, create a symlink to the
+         # “fat manifest” with the conventional name for a skinny manifest.
+         # This works because the file we just saved as the “fat manifest” is
+         # actually a misleadingly named skinny manifest. Link is relative to
+         # avoid embedding the storage directory path within the storage
+         # directory (see PR #1657).
+         if (not self.manifest_path.exists_()):
+            self.manifest_path.symlink_to(self.fatman_path.name)
          raise ch.No_Fatman_Error()
       if ("errors" in fm):
          # fm is an error blob.
@@ -209,7 +219,7 @@ class Image_Puller:
       "Return the path to tarball for layer layer_hash."
       return ch.storage.download_cache // (layer_hash + ".tar.gz")
 
-   def manifest_load(self):
+   def manifest_load(self, have_skinny=False):
       """Download the manifest file, parse it, and set self.config_hash and
          self.layer_hashes. If the image does not exist,
          exit with error."""
@@ -219,7 +229,7 @@ class Image_Puller:
       self.layer_hashes = None
       # obtain the manifest
       try:
-         # internal manifest library, e.g. for "FROM scratch"
+         # internal manifest library, e.g. for “FROM scratch”
          manifest = manifests_internal[str(self.src_ref)]
          ch.INFO("manifest: using internal library")
       except KeyError:
@@ -229,9 +239,10 @@ class Image_Puller:
          else:
             digest = self.architectures[ch.arch]
          ch.DEBUG("manifest digest: %s" % digest)
-         self.registry.manifest_to_file(self.manifest_path,
-                                        "manifest: downloading",
-                                        digest=digest)
+         if (not have_skinny):
+            self.registry.manifest_to_file(self.manifest_path,
+                                          "manifest: downloading",
+                                          digest=digest)
          manifest = self.manifest_path.json_from_file("manifest")
       # validate schema version
       try:
@@ -244,7 +255,7 @@ class Image_Puller:
       #
       # FIXME: Manifest version 1 does not list a config blob. It does have
       # things (plural) that look like a config at history/v1Compatibility as
-      # an embedded JSON string :P but I haven't dug into it.
+      # an embedded JSON string :P but I haven’t dug into it.
       if (version == 1):
          ch.VERBOSE("no config; manifest schema version 1")
          self.config_hash = None
@@ -271,7 +282,7 @@ class Image_Puller:
          self.layer_hashes.append(ch.digest_trim(i[key2]))
       if (version == 1):
          self.layer_hashes.reverse()
-      # Remember State_ID input. We can't rely on the manifest existing in
+      # Remember State_ID input. We can’t rely on the manifest existing in
       # serialized form (e.g. for internal manifests), so re-serialize.
       self.sid_input = json.dumps(manifest, sort_keys=True)
 
